from __future__ import annotations
from argparse import ArgumentParser
from multiprocessing import cpu_count
from pathlib import Path

from PIL import Image

from svgbit import STDataset
from svgbit import run
from svgbit import svg_heatmap


def main() -> None:
    parser = ArgumentParser(
        prog="svgbit",
        description="Find spatial variable genes for Spatial Trasncriptomics data.",
    )
    parser.add_argument(
        "count",
        help="path to count matrix file (shape: (spot * gene))",
    )
    parser.add_argument(
        "coordinate",
        help="path to coordinate file (shape: (spot * 2))",
    )
    parser.add_argument(
        "--count_transpose",
        action="store_true",
        help="transpose count matrix if specified",
    )
    parser.add_argument(
        "--coordinate_transpose",
        action="store_true",
        help="transpose coordinate file if specified",
    )
    parser.add_argument(
        "--k",
        type=int,
        default=6,
        help="number of nearest neighbors for KNN network (default: %(default)s)",
    )
    parser.add_argument(
        "--n_svgs",
        type=int,
        default=1000,
        help="number of SVGs to find clusters (default: %(default)s)",
    )
    parser.add_argument(
        "--n_svg_clusters",
        type=int,
        default=8,
        help="number of SVG clusters to find (default: %(default)s)",
    )
    parser.add_argument(
        "--he_image",
        default=None,
        help="path to H&E image. Only used for visualization (default: %(default)s)",
    )
    parser.add_argument(
        "--savedir",
        default=".",
        help="path to save results (default: %(default)s)",
    )
    parser.add_argument(
        "--cores",
        type=int,
        default=cpu_count(),
        help="number of threads to run svgbit (default: %(default)s)",
    )
    args = parser.parse_args()

    d = STDataset(
        args.count,
        args.coordinate,
        count_transpose=args.count_transpose,
        coordinate_transpose=args.coordinate_transpose,
        count_df_kwargs={"index_col": 0, "header": 0},
        coordinate_df_kwargs={"index_col": 0, "header": 0},
    )
    d = run(
        d,
        k=args.k,
        n_svgs=args.n_svgs,
        n_svg_clusters=args.n_svg_clusters,
        cores=args.cores,
    )

    savedir = Path(args.savedir)
    d.hotspot_df.to_csv(Path.joinpath(savedir, "hotspot_df.csv"))
    d.AI.to_csv(Path.joinpath(savedir, "AI.csv"))
    d.Di.to_csv(Path.joinpath(savedir, "Di.csv"))
    d.svg_cluster.to_csv(Path.joinpath(savedir, "svg_cluster.csv"))

    he_image = None if args.he_image is None else Image.open(args.he_image)
    svg_heatmap(d, Path.joinpath(savedir, "heatmap.jpg"), he_image)
    None if he_image is None else he_image.close()


if __name__ == "__main__":
    main()



from __future__ import annotations

import sys
import warnings
from typing import Optional, Tuple, Union
from pathlib import Path

import matplotlib as mpl
import numpy as np
import pandas as pd
from libpysal.weights import KNN
from libpysal.weights import W as libpysal_W

from . import cluster, density, moran, plot

DataFrames = Union[pd.DataFrame, np.ndarray, Path, str]
if not sys.warnoptions:
    import os
    warnings.simplefilter("default")  # Change the filter in this process
    os.environ["PYTHONWARNINGS"] = "default"  # Also affect subprocesses


class STDataset(object):
    """
    STDataset: A meta class for discribing Spatial Transcriptomics data.

    Parameters
    ==========
    count_df : np.ndarray, pd.DataFrame, str or Path
       Expression matrix for Spatial Transcriptomics Data. If ``str`` or ``Path``
       is given, svgbit will try to read file with given path with pandas.

       Default shape: (spot * gene)

    coordinate_df : np.ndarray, pd.DataFrame, str or Path
       Coordinates for Spatial Transcriptomics Data. If ``str`` or ``Path``
       is given, svgbit will try to read file with given path with pandas.

       Default shape: (spot * 2)

    count_transpose : bool, default False
        Whether to transpose count matrix.

    coordinate_transpose : bool, default False
        Whether to transpose coordinate dataframe.

    count_df_kwargs : dict, default {}
        Keyword arguments pass to ``pandas.read_csv`` if ``str`` or ``Path`` is
        given to ``count_df``.

    coordinate_df_kwargs : dict, default {}
        Keyword arguments pass to ``pandas.read_csv`` if ``str`` or ``Path`` is
        given to ``coordinate_df``.

    """
    def __init__(
        self,
        count_df: DataFrames,
        coordinate_df: DataFrames,
        count_transpose: bool = False,
        coordinate_transpose: bool = False,
        count_df_kwargs: dict = {},
        coordinate_df_kwargs: dict = {},
    ) -> None:

        # attributes initial
        self._count_df: Optional[pd.DataFrame] = None
        self._coordinate_df: Optional[pd.DataFrame] = None
        self._weight: Optional[libpysal_W] = None
        self._weight_type: Tuple[Optional[str], Optional[str]] = (None, None)
        self._hotspot_df: Optional[pd.DataFrame] = None
        self._local_moran_i: Optional[pd.DataFrame] = None
        self._local_moran_p: Optional[pd.DataFrame] = None
        self._AI: Optional[pd.DataFrame] = None
        self._Di: Optional[pd.DataFrame] = None
        self._svg_cluster: Optional[pd.Series] = None

        # dataframes check
        if isinstance(count_df, pd.DataFrame):
            self._count_df = count_df
        elif isinstance(count_df, np.ndarray):
            self._count_df = pd.DataFrame(count_df)
        else:
            self._count_df = pd.read_csv(count_df, **count_df_kwargs)
        if count_transpose:
            self._count_df = self._count_df.T

        if isinstance(coordinate_df, pd.DataFrame):
            self._coordinate_df = coordinate_df
        elif isinstance(coordinate_df, np.ndarray):
            self._coordinate_df = pd.DataFrame(coordinate_df)
        else:
            self._coordinate_df = pd.read_csv(
                coordinate_df,
                **coordinate_df_kwargs,
            )
        if coordinate_transpose:
            self._coordinate_df = self._coordinate_df.T

        self._count_df.sort_index(inplace=True)
        self._coordinate_df.sort_index(inplace=True)
        self._coordinate_df.columns = ["X", "Y"]

        err = "Expression matrix and coordinate file have different number of spots."
        assert self._count_df.shape[0] == self._coordinate_df.shape[0], err
        err = "Spots' name mismatch!"
        assert all(self._count_df.index == self._coordinate_df.index), err

        # Rename duplicated columns
        cols = pd.Series(self._count_df.columns)
        dups = self._count_df.columns[self._count_df.columns.duplicated(
            keep=False)]
        for dup in dups:
            cols[self._count_df.columns.get_loc(dup)] = ([
                dup + '.' + str(d_idx) if d_idx != 0 else dup
                for d_idx in range(self._count_df.columns.get_loc(dup).sum())
            ])
        if len(dups) > 0:
            warnings.warn("Duplicated column names found. Auto rename.")
            self._count_df.columns = cols

    def acquire_weight(self, k: int = 6, **kwargs) -> None:
        """
        Acquire weight for analysis.

        Parameters
        ==========
        k: int, default 6
            Number of nearest neighbors for KNN network.

        **kwargs
            Additional keyword arguments passed to the libpysal.weights.KNN call.

        """
        self._weight = KNN(self._coordinate_df, k=k, **kwargs)
        self._weight_type = ("KNN", str(k))

    def acquire_hotspot(self, **kwargs) -> None:
        """
        Acquire hotspot matrix.

        Parameters
        ==========
        **kwargs
            Additional keyword arguments passed to local_moran call.

        """
        if self._weight is None:
            self.acquire_weight()
        hotspot, i_value, p_value = moran.local_moran(
            gene_expression_df=self.count_df,
            weights=self.weight,
            **kwargs,
        )
        self._hotspot_df = hotspot.reindex(
            index=self.spots,
            columns=self.genes,
        )
        self._local_moran_i = i_value
        self._local_moran_p = p_value

    def acquire_density(self, cores: int = density.cpu_count()) -> None:
        """
        Acquire local Di and global AI value.

        Parameters
        ==========
        cores: int
            Number of threads to run svgbit. Use all available cpus by default.

        """
        if self._hotspot_df is None:
            self.acquire_hotspot()
        self._AI, self._Di = density.hotspot_AI(
            hotspot_df=self._hotspot_df,
            weight_df=self._local_moran_p,
            knn=self._weight,
            cores=cores,
        )

    def find_clusters(
        self,
        n_svgs: int = 1000,
        n_svg_clusters: int = 8,
    ) -> None:
        """
        Find SVG clusters.

        Parameters
        ==========
        n_svgs: int, default 1000
            Number of SVGs to find clusters.

        n_svg_clusters: int, default 8
            Number of SVG clusters to find.

        """
        self._svg_cluster = cluster.cluster(
            self._hotspot_df,
            self._AI,
            n_svgs=n_svgs,
            n_svg_clusters=n_svg_clusters,
        )

    def svg_heatmap(
        self,
        save_path: Union[str, Path],
        he_image=None,
    ) -> Tuple[mpl.figure.Figure, np.ndarray[mpl.axes.Axes]]:
        """
        Draw SVG distribution heatmap.

        Patameters
        ==========
        save_path: str or pathlib.Path
            Heatmap save path.

        he_image: PIL.Image.Image, default None
            H&E image of tissue. If None is given (default), distribution map
            will not show tissue picture.
        """
        plot.svg_heatmap(
            self._hotspot_df,
            self._coordinate_df,
            self._svg_cluster,
            save_path,
            he_image,
        )

    @property
    def count_df(self) -> pd.DataFrame:
        """Expression matrix."""
        return self._count_df

    @property
    def coordinate_df(self) -> pd.DataFrame:
        """Coordinate information."""
        return self._coordinate_df

    @property
    def n_spots(self) -> int:
        """Number of total spots."""
        return self._count_df.shape[0]

    @property
    def spots(self) -> pd.Index:
        """An Index for spots' names."""
        return self._count_df.index

    @property
    def n_genes(self) -> int:
        """Number of total genes."""
        return self._count_df.shape[1]

    @property
    def genes(self) -> pd.Index:
        """An Index for genes' names."""
        return self._count_df.columns

    @property
    def weight(self) -> libpysal_W:
        """Weight used by svgbit. Use KNN if not specified."""
        return self._weight

    @weight.setter
    def weight(self, value: libpysal_W):
        self._weight = value
        self._weight_type = ("User specified weight", None)

    @property
    def weight_type(self) -> Tuple[Optional[str], Optional[str]]:
        """
        What kind of weight is used. The second element indicates parameter k
        used by KNN in default.
        """
        return self._weight_type

    @property
    def hotspot_df(self) -> pd.DataFrame:
        """Hotspot matrix."""
        return self._hotspot_df

    @property
    def AI(self) -> pd.Series:
        """A Series for AI value."""
        return self._AI

    @property
    def Di(self) -> pd.DataFrame:
        """A DataFrame for local Di value."""
        return self._Di

    @property
    def svg_cluster(self) -> pd.Series:
        """SVG cluster result."""
        return self._svg_cluster


if __name__ == "__main__":
    pass



from __future__ import annotations

from functools import partial
from multiprocessing import Pool, cpu_count
from typing import Tuple

import numba
import pandas as pd
from libpysal.weights import W as libpysal_W
from pysal.explore import esda


def local_moran(
        gene_expression_df: pd.DataFrame,
        weights: libpysal_W,
        transformation: str = 'r',
        permutation: int = 999,
        cores: int = cpu_count(),
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Calculate local moran for hotspot identification.

    Parameters
    ==========
    gene_expression_df : pd.DataFrame
       Expression matrix for Spatial Transcriptomics Data.

    weights : libpysal.weights.W
        Spatial weight for calculating Moran's I.

    transformation : str, default 'r'
        Weights transformation passed to esda.moran.Moran_Local

    permutation : int, default 999
        Number of random permutations for calculation of pseudo-p_values.

    cores : int
        Number of threads to run svgbit. Use all available cpus by default.

    Returns
    =======
    hotspot : pd.DataFrame
        Identified hotspots.

    i_value : pd.DataFrame
        Local Moran's I values.

    p_value : pd.DataFrame
        Local Moran's I p values.

    """
    partial_func = partial(
        _local_moran,
        gene_expression_df=gene_expression_df,
        weights=weights,
        transformation=transformation,
        permutation=permutation,
    )
    pool = Pool(processes=cores)
    result_lists = pool.map(partial_func, gene_expression_df.columns)
    pool.close()
    pool.join()
    hotspot = pd.concat([i[0] for i in result_lists], axis=1)
    i_value = pd.concat([i[1] for i in result_lists], axis=1)
    p_value = pd.concat([i[2] for i in result_lists], axis=1)

    return hotspot, i_value, p_value


def _local_moran(
    gene: str,
    gene_expression_df: pd.DataFrame,
    weights: libpysal_W,
    transformation: str = "r",
    permutation: int = 999,
) -> Tuple[pd.Series, pd.Series, pd.Series]:
    """
    Calculate local moran for hotspot identification for a single gene.

    Parameters
    ==========
    gene : str
        Calculate which gene.

    gene_expression_df : pd.DataFrame
       Expression matrix for Spatial Transcriptomics Data.

    weights : libpysal.weights.W
        Spatial weight for calculating Moran's I.

    transformation : str, default 'r'
        Weights transformation passed to esda.moran.Moran_Local

    permutation : int, default 999
        Number of random permutations for calculation of pseudo-p_values.

    Returns
    =======
    hotspot : pd.DataFrame
        Identified hotspots.

    i_value : pd.DataFrame
        Local Moran's I values.

    p_value : pd.DataFrame
        Local Moran's I p values.

    """
    numba.config.THREADING_LAYER = "workqueue"
    gene_lisa = esda.moran.Moran_Local(
        gene_expression_df[gene],
        weights,
        transformation=transformation,
        permutations=permutation,
        n_jobs=1,
    )
    # select the significant hotspots (p < 0.05)
    local_moran_hotspot = 1 * ((gene_lisa.p_sim < 0.05) & (gene_lisa.q == 1))

    hotspot = pd.Series(
        local_moran_hotspot,
        index=gene_expression_df.index,
        name=gene,
    )
    if (hotspot == 1).all():
        hotspot = 0

    i_value = pd.Series(
        gene_lisa.EI_sim,
        index=gene_expression_df.index,
        name=gene,
    )
    p_value = pd.Series(
        gene_lisa.p_sim,
        index=gene_expression_df.index,
        name=gene,
    )

    return hotspot, i_value, p_value


def global_moran(
        gene_expression_df: pd.DataFrame,
        weights: libpysal_W,
        transformation: str = 'r',
        permutation: int = 999,
        cores: int = cpu_count(),
) -> pd.DataFrame:
    """
    Calculate global Moran's I value.

    Parameters
    ==========
    gene_expression_df : pd.DataFrame
       Expression matrix for Spatial Transcriptomics Data.

    weights : libpysal.weights.W
        Spatial weight for calculating Moran's I.

    transformation : str, default 'r'
        Weights transformation passed to esda.moran.Moran

    permutation : int, default 999
        Number of random permutations for calculation of pseudo-p_values.

    cores : int
        Number of threads to run svgbit. Use all available cpus by default.

    Returns
    =======
    global_moran_df : pd.DataFrame
        Global Moran's I result.

    """
    partial_func = partial(
        _global_moran,
        gene_expression_df=gene_expression_df,
        weights=weights,
        transformation=transformation,
        permutation=permutation,
    )
    pool = Pool(processes=cores)
    result_lists = pool.map(partial_func, gene_expression_df.columns)
    pool.close()
    pool.join()
    result_df = pd.concat(result_lists, axis=0)

    global_moran_df = result_df
    global_moran_df.sort_values(
        by="p_value_sim",
        inplace=False,
        ascending=True,
    )

    return global_moran_df


def _global_moran(
    gene: str,
    gene_expression_df: pd.DataFrame,
    weights: libpysal_W,
    transformation: str = 'r',
    permutation: int = 999,
) -> pd.DataFrame:
    """
    Calculate global Moran's I value for a single gene.

    Parameters
    ==========
    gene : str
        Calculate which gene.

    gene_expression_df : pd.DataFrame
       Expression matrix for Spatial Transcriptomics Data.

    weights : libpysal.weights.W
        Spatial weight for calculating Moran's I.

    transformation : str, default 'r'
        Weights transformation passed to esda.moran.Moran

    permutation : int, default 999
        Number of random permutations for calculation of pseudo-p_values.

    Returns
    =======
    global_moran_df : pd.DataFrame
        Global Moran's I result.

    """
    global_moran_df = pd.DataFrame(
        columns=[
            'I_value',
            'p_value_sim',
            'p_value_rand',
            'p_value_z_sim',
            'p_value_norm',
        ],
        index=[gene],
    )
    moran_value = esda.moran.Moran(
        gene_expression_df[gene],
        weights,
        transformation=transformation,
        permutations=permutation,
    )
    p_value_sim = moran_value.p_sim
    moranI = moran_value.I
    p_value_norm = moran_value.p_norm
    p_value_rand = moran_value.p_rand
    p_value_z_sim = moran_value.p_z_sim
    global_moran_df['p_value_sim'][gene] = p_value_sim
    global_moran_df['I_value'][gene] = moranI
    global_moran_df['p_value_rand'][gene] = p_value_rand
    global_moran_df['p_value_z_sim'] = p_value_z_sim
    global_moran_df['p_value_norm'] = p_value_norm

    return global_moran_df


if __name__ == "__main__":
    pass



from __future__ import annotations

from functools import partial
from multiprocessing import Pool, cpu_count
from typing import Tuple, Union

import numpy as np
import pandas as pd
from libpysal.weights import W as libpysal_W

from .utils import pysal_to_pandas


def _hotspot_AI(
    gene: str,
    hotspot_df: pd.DataFrame,
    weight_df: pd.DataFrame,
    knn: Union[pd.DataFrame, libpysal_W],
) -> Tuple[pd.Series, pd.Series]:
    """
    Calculate Spatial Transcriptomics AI value for a single gene.

    Parameters
    ==========
    gene : str
        Calculate which gene.

    hotspot_df : pd.DataFrame
        A hotspot DataFrame generated by svgbit.

    weight_df : pd.DataFrame
        Weight used by AI. In default, svgbit uses local Moran's I p value as weight.

    knn : pd.DataFrame or libpysal.weights.W
        KNN network used for neighbor idenfication.

    Returns
    =======
    ai_series : pd.Series
        A Series for AI value.

    di_series : pd.Series
        A Series for local Di value.

    """
    if isinstance(knn, libpysal_W):
        knn = pysal_to_pandas(knn)
        knn.index = hotspot_df.index
        knn.columns = hotspot_df.index

    gene_hotspot_df = hotspot_df[hotspot_df[gene] != 0]
    gene_hotspot_df = pd.DataFrame(gene_hotspot_df[gene], columns=[gene])
    n_hotspots = len(gene_hotspot_df)
    if n_hotspots == 0:
        ai_series = pd.Series([0], index=[gene], name="AI")
        di_series = pd.Series(
            [0] * len(hotspot_df),
            index=hotspot_df.index,
            name=gene,
        )
        return ai_series, di_series
    hotspots = gene_hotspot_df.index.tolist()
    hs = {}
    for i in hotspots:
        hi_wnn_df = pd.DataFrame(knn[i], columns=[i])
        hi_wnn_df = pd.DataFrame(hi_wnn_df[hi_wnn_df[i] == 1])
        knn_coors = hi_wnn_df.index.tolist()
        inter_spots = set(hotspots).intersection(set(knn_coors))
        hs[i] = ((weight_df[gene][inter_spots] /
                  weight_df[gene][knn_coors].sum()).sum())

    di_array = np.array(list(hs.values()))
    di_sum = di_array.sum()
    di_mean = di_sum / n_hotspots
    ai_series = pd.Series([di_mean], index=[gene], name="AI")
    di_series = pd.Series(
        hs,
        name=gene,
    ).reindex(index=hotspot_df.index).fillna(0)

    return ai_series, di_series


def hotspot_AI(
        hotspot_df: pd.DataFrame,
        weight_df: pd.DataFrame,
        knn: Union[pd.DataFrame, libpysal_W],
        cores: int = cpu_count(),
) -> Tuple[pd.Series, pd.DataFrame]:
    """
    Calculate Spatial Transcriptomics AI value for all genes.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A hotspot DataFrame generated by svgbit.

    weight_df : pd.DataFrame
        Weight used by AI. In default, svgbit uses local Moran's I p value as weight.

    knn : pd.DataFrame or libpysal.weights.W
        KNN network used for neighbor idenfication.

    cores : int
        Number of threads to run svgbit. Use all available cpus by default.

    Returns
    =======
    ai_series : pd.Series
        A Series for AI value.

    di_df : pd.DataFrame
        A DataFrame for local Di value.

    """
    if isinstance(knn, libpysal_W):
        knn = pysal_to_pandas(knn)
        knn.index = hotspot_df.index
        knn.columns = hotspot_df.index
    partial_func = partial(
        _hotspot_AI,
        hotspot_df=hotspot_df,
        weight_df=weight_df,
        knn=knn,
    )
    pool = Pool(processes=cores)
    result_lists = pool.map(partial_func, hotspot_df.columns)
    pool.close()
    pool.join()
    ai_series = pd.concat([i[0] for i in result_lists], axis=0)
    di_df = pd.concat([i[1] for i in result_lists], axis=1)

    return ai_series, di_df


if __name__ == "__main__":
    pass



from __future__ import annotations

from multiprocessing import cpu_count
from pathlib import Path
from typing import Optional, Union

from PIL import Image

from .core.STDataset import STDataset
from .core.io import load_10X
from .core.plot import _svg_heatmap


def run(
        dataset: STDataset,
        k: int = 6,
        n_svgs: int = 1000,
        n_svg_clusters: int = 8,
        cores: int = cpu_count(),
) -> STDataset:
    """
    Run all part of svgbit within one function.

    Parameters
    ==========
    dataset : STDataset
        A STDataset for running svgbit.

    k: int, default 6
        Number of nearest neighbors for KNN network.

    n_svgs: int, default 1000
        Number of SVGs to find clusters.

    n_svg_clusters: int, default 8
        Number of SVG clusters to find.

    cores: int
        Number of threads to run svgbit. Use all available cpus by default.

    Returns
    =======
    dataset : STDataset
        A STDataset with all evaluates done.

    """
    dataset.acquire_weight(k=k)
    dataset.acquire_hotspot(cores=cores)
    dataset.acquire_density(cores=cores)
    dataset.find_clusters(n_svgs=n_svgs, n_svg_clusters=n_svg_clusters)

    return dataset


def svg_heatmap(
    dataset: STDataset,
    save_path: Union[str, Path],
    he_image: Optional[Image.Image] = None,
) -> None:
    """
    Draw SVG distribution heatmap.

    Patameters
    ==========
    dataset: STDataset
        A STDataset with hotspot and SVG cluster estimation finished.

    save_path: str or pathlib.Path
        Heatmap save path.

    he_image: PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.
    """
    _svg_heatmap(
        dataset.hotspot_df,
        dataset.coordinate_df,
        dataset.svg_cluster,
        save_path,
        he_image,
    )



from __future__ import annotations

import pandas as pd
from libpysal.weights import W as libpysal_W


def pysal_to_pandas(W: libpysal_W) -> pd.DataFrame:
    return pd.DataFrame(*W.full()).astype(int).T



from __future__ import annotations

import gzip
from pathlib import Path

import pandas as pd
import scipy.io
from .STDataset import STDataset


def load_10X(read_dir) -> STDataset:
    """
    Load 10X Genomics Space Ranger outputs and generate STDataset.

    Patameters
    ==========
    read_dir: str or pathlib.Path
        A location points to 10X outs dir. Assume the directories
        ``filtered_feature_bc_matrix`` and ``spatial`` are in this
        path.

    Returns
    =======
    dataset: STDataset
        A STDataset instance generated from read_dir.
    """
    read_dir = Path(read_dir)
    mat_dir = Path.joinpath(read_dir, "filtered_feature_bc_matrix")
    mtx_path = Path.joinpath(mat_dir, "matrix.mtx.gz")
    features_path = Path.joinpath(mat_dir, "features.tsv.gz")
    barcodes_path = Path.joinpath(mat_dir, "barcodes.tsv.gz")
    position_path = Path.joinpath(read_dir, "spatial",
                                  "tissue_positions_list.csv")

    gene_name = []
    with gzip.open(features_path, "rt") as f:
        for line in f:
            line = line.strip()
            gene_name.append(line.split("\t")[1])

    spot_name = []
    with gzip.open(barcodes_path, "rt") as f:
        for line in f:
            line = line.strip()
            spot_name.append(line)

    count_df = pd.DataFrame.sparse.from_spmatrix(
        scipy.io.mmread(mtx_path),
        index=gene_name,
        columns=spot_name,
    ).T
    coor_df = pd.read_csv(position_path, index_col=0, header=None)
    coor_df = coor_df[[5, 4]]
    coor_df.index.name = "barcode"
    coor_df.columns = ["X", "Y"]
    coor_df = coor_df.reindex(index=count_df.index)

    dataset = STDataset(count_df, coor_df)
    return dataset



import re
from setuptools import setup, find_packages

VERSIONFILE = "svgbit/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." %
                       (VERSIONFILE, ))

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="svgbit",
    version=verstr,
    author="CPenglab",
    author_email="chengpeng@ynu.edu.cn",
    discription="Find spatial variable genes for Spatial Trasncriptomics data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CPenglab/svgbit",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "setuptools",
        "matplotlib>=3.5.1",
        "pandana>=0.6.1",
        "pysal>=2.4.0",
        "pillow>=9.2.0",
    ],
    entry_points={'console_scripts': [
        'svgbit = svgbit.__main__:main',
    ]},
    python_requires=">=3.8.12",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ])



from __future__ import annotations

import pandas as pd
from scipy.cluster import hierarchy as sch


def cluster(
    hotspot_df: pd.DataFrame,
    AI_series: pd.Series,
    n_svgs: int = 1000,
    n_svg_clusters: int = 8,
) -> pd.Series:
    """
    Cluster SVGs using hotspot matrix.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A hotspot DataFrame generated by svgbit.

    AI_series : pd.Series
        A Series for AI value.

    n_svgs : int, default 1000
        Number of SVGs to find clusters.

    n_svg_clusters : int, default 8
        Number of SVG clusters to find.

    Returns
    =======
    gene_result : pd.Series
        A Series of clustering result.

    """
    selected_genes = AI_series.sort_values(ascending=False)[:n_svgs].index
    hotspot_set = hotspot_df[selected_genes]
    gene_distmat = sch.distance.pdist(hotspot_set.T, metric="jaccard")
    Z_gene = sch.linkage(gene_distmat, method="ward")
    gene_result = pd.Series(
        sch.fcluster(Z_gene, t=n_svg_clusters, criterion="maxclust"),
        index=selected_genes,
    ).sort_values()
    return gene_result



from __future__ import annotations

from math import ceil
from pathlib import Path
from typing import Union

import matplotlib.pyplot as plt
import pandas as pd
from PIL import Image
from scipy.cluster import hierarchy as sch


def _svg_heatmap(
    hotspot_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    cluster_result: pd.Series,
    save_path: Union[str, Path],
    he_image: Image.Image = None,
) -> None:
    """
    Draw SVG distribution heatmap.

    Patameters
    ==========
    hotspot_df: pd.DataFrame
        A pd.DataFrame for hotspots.

    coordinate_df: pd.DataFrame
        A pd.DataFrame for coordinate files.

    cluster_result: pd.Series
        A pd.Series for cluster result.

    save_path: str or pathlib.Path
        Heatmap save path.

    he_image: PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.
    """

    left, bottom = 0.1, 0.1
    width, height = 0.66, 0.66
    spacing, cluster_width = 0.03, 0.22
    rect_heatmap = [left + spacing, bottom + spacing * 2, width, height]

    fig = plt.figure(figsize=(10, 10), dpi=300)
    ax_heatmap = fig.add_axes(rect_heatmap)
    axes = [ax_heatmap]

    for i, j in enumerate(set(cluster_result.values)):
        rect_cluster = [
            left + width + spacing * (2 + i // 3) + cluster_width * (i // 3),
            bottom + spacing * (3 - i % 3) + cluster_width * (2 - i % 3),
            cluster_width,
            cluster_width,
        ]
        ax_cluster = fig.add_axes(rect_cluster)
        ax_cluster.set_title(f"Cluster {j} hotspot distribution")
        ax_cluster.set_xticks([])
        ax_cluster.set_yticks([])
        ax_cluster.imshow(he_image) if he_image is not None else None
        None if he_image is None else ax_cluster.imshow(he_image)
        ax_cluster.axis("off")
        sc = ax_cluster.scatter(
            coordinate_df.iloc[:, 0],
            coordinate_df.iloc[:, 1],
            c=hotspot_df[cluster_result[cluster_result == j].index].T.mean(),
            cmap="autumn_r",
            vmin=0,
            vmax=1,
            s=4,
            alpha=0.7,
        )
        axes.append(ax_cluster)

    ii = ceil(len(set(cluster_result.values)) / 3)
    rect_cb = [
        left + width + spacing * (ii + 2) + cluster_width * ii,
        bottom + spacing * 2,
        0.02,
        height,
    ]
    ax_cb = fig.add_axes(rect_cb)
    fig.colorbar(sc, cax=ax_cb)
    axes.append(ax_cb)

    spot_distmat = sch.distance.pdist(hotspot_df, metric="jaccard")
    Z_spot = sch.linkage(spot_distmat, method="ward")
    spot_result = pd.Series(
        sch.fcluster(Z_spot, t=8, criterion="maxclust"),
        index=hotspot_df.index,
    ).sort_values()

    ax_heatmap.imshow(
        hotspot_df.reindex(
            index=spot_result.index,
            columns=cluster_result.index,
        ),
        cmap="Reds",
        aspect="auto",
    )

    flag = 0
    for i in range(1, len(set(cluster_result.values))):
        flag += len(cluster_result[cluster_result == i])
        ax_heatmap.plot([flag, flag], [0, hotspot_df.shape[0] - 5])

    ax_heatmap.set_xticks([])
    ax_heatmap.set_yticks([])
    ax_heatmap.set_xlabel("Genes")
    ax_heatmap.set_ylabel("Spots")

    fig.savefig(save_path, bbox_inches="tight")
